@@ -4,6 +4,7 @@
 #include <time.h>
 
 #include <sstream>
+#include <iostream>
 
 #ifndef INT32_MAX
 #define INT32_MAX 0x7fffffff
@@ -17,6 +18,8 @@ using namespace std;
 
 int V8Context::number = 0;
 
+v8::Isolate* isolate;
+
 void set_perl_error(const TryCatch& try_catch) {
     Handle<Message> msg = try_catch.Message();
 
@@ -26,7 +29,7 @@ void set_perl_error(const TryCatch& try_catch) {
         1024,
         "%s at %s:%d:%d\n",
         *(String::Utf8Value(try_catch.Exception())),
-        !msg.IsEmpty() ? *(String::AsciiValue(msg->GetScriptResourceName())) : "eval",
+        !msg.IsEmpty() ? *(String::Utf8Value(msg->GetScriptResourceName())) : "eval",
         !msg.IsEmpty() ? msg->GetLineNumber() : 0,
         !msg.IsEmpty() ? msg->GetStartColumn(): 0
     );
@@ -43,9 +46,9 @@ check_perl_error() {
     const char *err = SvPV_nolen(ERRSV);
 
     if (err && strlen(err) > 0) {
-        Handle<String> error = String::New(err, strlen(err) - 1); // no newline
+        Handle<String> error = v8::String::NewFromUtf8(isolate, err, v8::String::kNormalString);
         sv_setsv(ERRSV, &PL_sv_no);
-        Handle<Value> v = ThrowException(Exception::Error(error));
+        Handle<Value> v = isolate->ThrowException(Exception::Error(error));
         return v;
     }
 
@@ -129,7 +132,7 @@ void SvMap::add(Handle<Object> object, long ptr) {
 SV* SvMap::find(Handle<Object> object) {
     int hash = object->GetIdentityHash();
 
-    for (sv_map::const_iterator it = objects.find(hash); it != objects.end(), it->first == hash; it++)
+    for (sv_map::const_iterator it = objects.find(hash); it != objects.end() && it->first == hash; it++)
         if (it->second->object->Equals(object))
             return newRV_inc(INT2PTR(SV*, it->second->ptr));
 
@@ -137,10 +140,14 @@ SV* SvMap::find(Handle<Object> object) {
 }
 
 ObjectData::ObjectData(V8Context* context_, Handle<Object> object_, SV* sv_)
-    : context(context_)
-    , object(Persistent<Object>::New(object_))
-    , sv(sv_)
 {
+    context = context_;
+    sv = sv_;
+
+    Persistent<Object, CopyablePersistentTraits<Object>> persistent(isolate, object_);
+
+    object = persistent;
+
     if (!sv) return;
 
     ptr = PTR2IV(sv);
@@ -150,7 +157,7 @@ ObjectData::ObjectData(V8Context* context_, Handle<Object> object_, SV* sv_)
 
 ObjectData::~ObjectData() {
     if (context) context->remove_object(this);
-    object.Dispose();
+    object.Reset();
 }
 
 PerlObjectData::PerlObjectData(V8Context* context_, Handle<Object> object_, SV* sv_)
@@ -164,7 +171,7 @@ PerlObjectData::PerlObjectData(V8Context* context_, Handle<Object> object_, SV*
     add_size(calculate_size(sv));
     ptr = PTR2IV(sv);
 
-    object.MakeWeak(this, PerlObjectData::destroy);
+    object.SetWeak(this, PerlObjectData::destroy, v8::WeakCallbackType::kParameter);
 }
 
 size_t PerlObjectData::size() {
@@ -197,8 +204,9 @@ int V8ObjectData::svt_free(pTHX_ SV* sv, MAGIC* mg) {
     return 0;
 };
 
-void PerlObjectData::destroy(Persistent<Value> object, void *data) {
-    delete static_cast<PerlObjectData*>(data);
+void PerlObjectData::destroy(const WeakCallbackInfo<PerlObjectData>& data) {
+    data.GetParameter()->object.Reset();
+    delete static_cast<PerlObjectData*>(data.GetParameter());
 }
 
 ObjectData* sv_object_data(SV* sv) {
@@ -214,7 +222,7 @@ class V8FunctionData : public V8ObjectData {
 public:
     V8FunctionData(V8Context* context_, Handle<Object> object_, SV* sv_)
         : V8ObjectData(context_, object_, sv_)
-        , returns_list(object_->Has(String::New("__perlReturnsList")))
+        , returns_list(object_->Has(String::NewFromUtf8(isolate, "__perlReturnsList", v8::String::kNormalString)))
     { }
 
     bool returns_list;
@@ -226,7 +234,7 @@ private:
     Local<Value> *thisWrapped;
 
 protected:
-    virtual Handle<Value> invoke(const Arguments& args);
+    virtual Handle<Value> invoke(const v8::FunctionCallbackInfo<v8::Value>& args);
     virtual size_t size();
 
 public:
@@ -234,24 +242,30 @@ public:
         : PerlObjectData(
               context_,
               Handle<Object>::Cast(
-                  context_->make_function->Call(
-                      context_->context->Global(),
+                  context_->make_function.Get(isolate)->Call(
+                      context_->get_local_context(),
+                      context_->get_local_context()->Global(),
                       1,
-                      thisWrapped = new Local<Value>(External::Wrap(this))
-                  )
+                      thisWrapped = new Local<Value>(External::New(isolate, this))
+                  ).ToLocalChecked()
               ),
               cv
           )
        , rv(cv ? newRV_noinc(cv) : NULL)
     { }
-    
+
     ~PerlFunctionData() {
         delete thisWrapped;
     }
 
-    static Handle<Value> v8invoke(const Arguments& args) {
-        PerlFunctionData* data = static_cast<PerlFunctionData*>(External::Unwrap(args[0]));
-        return data->invoke(args);
+    static void v8invoke(const v8::FunctionCallbackInfo<v8::Value>& args) {
+        v8::Local<v8::External> ext = args[0].As<v8::External>();
+
+        PerlFunctionData* data = static_cast<PerlFunctionData*>(ext->Value());
+
+        if (data != NULL) {
+            args.GetReturnValue().Set(data->invoke(args));
+        }
     }
 };
 
@@ -261,11 +275,11 @@ size_t PerlFunctionData::size() {
 
 void PerlObjectData::add_size(size_t bytes_) {
     bytes += bytes_;
-    V8::AdjustAmountOfExternalAllocatedMemory(bytes_);
+    //V8::AdjustAmountOfExternalAllocatedMemory(bytes_);
 }
 
 Handle<Value>
-PerlFunctionData::invoke(const Arguments& args) {
+PerlFunctionData::invoke(const v8::FunctionCallbackInfo<v8::Value>& args) {
     SETUP_PERL_CALL();
     int count = call_sv(rv, G_SCALAR | G_EVAL);
     CONVERT_PERL_RESULT();
@@ -274,7 +288,7 @@ PerlFunctionData::invoke(const Arguments& args) {
 class PerlMethodData : public PerlFunctionData {
 private:
     string name;
-    virtual Handle<Value> invoke(const Arguments& args);
+    virtual Handle<Value> invoke(const v8::FunctionCallbackInfo<v8::Value>& args);
     virtual size_t size();
 
 public:
@@ -285,7 +299,7 @@ public:
 };
 
 Handle<Value>
-PerlMethodData::invoke(const Arguments& args) {
+PerlMethodData::invoke(const v8::FunctionCallbackInfo<v8::Value>& args) {
     SETUP_PERL_CALL(mXPUSHs(context->v82sv(args.This())))
     int count = call_method(name.c_str(), G_SCALAR | G_EVAL);
     CONVERT_PERL_RESULT()
@@ -307,46 +321,99 @@ V8Context::V8Context(
       bless_prefix(bless_prefix_),
       enable_blessing(enable_blessing_)
 {
+    // Set flags before creating the isolate--otherwise some flags are
+    // ineffective.
     V8::SetFlagsFromString(flags, strlen(flags));
-    context = Context::New();
+
+    if (!isolate) {
+        //v8::V8::InitializeICU();
+        Platform* platform = platform::CreateDefaultPlatform();
+        V8::InitializePlatform(platform);
+        V8::Initialize();
+
+        v8::Isolate::CreateParams create_params;
+
+        create_params.array_buffer_allocator =
+            v8::ArrayBuffer::Allocator::NewDefaultAllocator();
+
+        isolate = v8::Isolate::New(create_params);
+    }
+
+    Isolate::Scope isolate_scope(isolate);
+
+    HandleScope handle_scope(isolate);
+
+    v8::Local<v8::ObjectTemplate> global = v8::ObjectTemplate::New(isolate);
+
+    v8::Local<v8::Context> context = Context::New(isolate, NULL, global);
+
+    Persistent<Context, CopyablePersistentTraits<Context>> persistent_context(isolate, context);
+
+    this->context = persistent_context;
 
     Context::Scope context_scope(context);
-    HandleScope handle_scope;
 
-    Local<FunctionTemplate> tmpl = FunctionTemplate::New(PerlFunctionData::v8invoke);
+    Local<FunctionTemplate> tmpl = FunctionTemplate::New(isolate, PerlFunctionData::v8invoke);
+
     context->Global()->Set(
-        String::New("__perlFunctionWrapper"),
+        v8::String::NewFromUtf8(isolate, "__perlFunctionWrapper", v8::String::kNormalString),
         tmpl->GetFunction()
     );
 
     Handle<Script> script = Script::Compile(
-        String::New(
+        v8::String::NewFromUtf8(
+            isolate,
             "(function(wrap) {"
             "    return function() {"
             "        var args = Array.prototype.slice.call(arguments);"
             "        args.unshift(wrap);"
             "        return __perlFunctionWrapper.apply(this, args)"
             "    };"
-            "})"
+            "})",
+            v8::String::kNormalString
         )
     );
-    make_function = Persistent<Function>::New(Handle<Function>::Cast(script->Run()));
 
-    string_wrap = Persistent<String>::New(String::New("wrap"));
+    make_function.Reset(isolate, Handle<Function>::Cast(script->Run()));
+
+    string_wrap.Reset(isolate, String::NewFromUtf8(isolate, "wrap"));
 
     number++;
 }
 
+Local<Context> V8Context::get_local_context() {
+    return Local<Context>::New(isolate, context);
+}
+
 void V8Context::register_object(ObjectData* data) {
     seen_perl[data->ptr] = data;
-    data->object->SetHiddenValue(string_wrap, External::Wrap(data));
+
+    v8::Local<v8::Private> privateKey = v8::Private::ForApi(isolate, string_wrap.Get(isolate));
+    auto local = v8::Local<v8::Object>::New(isolate, data->object);
+    v8::Local<v8::Context> context = isolate->GetCurrentContext();
+
+    local->SetPrivate(context, privateKey, External::New(isolate, data));
 }
 
 void V8Context::remove_object(ObjectData* data) {
     ObjectDataMap::iterator it = seen_perl.find(data->ptr);
     if (it != seen_perl.end())
         seen_perl.erase(it);
-    data->object->DeleteHiddenValue(string_wrap);
+
+    HandleScope scope(isolate);
+    Local<Context> local_context = Local<Context>::New(isolate, context);
+    Context::Scope context_scope(local_context);
+
+    v8::Local<v8::Private> privateKey = v8::Private::ForApi(isolate, string_wrap.Get(isolate));
+    auto local = v8::Local<v8::Object>::New(isolate, data->object);
+
+    if (local.IsEmpty()) {
+      return;
+    }
+
+    v8::Local<v8::Context> context = isolate->GetCurrentContext();
+
+    local->DeletePrivate(context, privateKey);
 }
 
 V8Context::~V8Context() {
@@ -356,37 +423,52 @@ V8Context::~V8Context() {
     seen_perl.clear();
 
     for (ObjectMap::iterator it = prototypes.begin(); it != prototypes.end(); it++) {
-      it->second.Dispose();
+      it->second.Reset();
     }
-    context.Dispose();
-    string_wrap.Dispose();
-    make_function.Dispose();
-    while(!V8::IdleNotification()); // force garbage collection
+    context.ClearWeak();
+    string_wrap.Reset();
+    make_function.Reset();
 }
 
 void
 V8Context::bind(const char *name, SV *thing) {
-    HandleScope scope;
-    Context::Scope context_scope(context);
+    Isolate::Scope isolate_scope(isolate);
+    HandleScope scope(isolate);
 
-    context->Global()->Set(String::New(name), sv2v8(thing));
+    Local<Context> local_context = Local<Context>::New(isolate, context);
+    Context::Scope context_scope(local_context);
+
+    local_context->Global()->Set(v8::String::NewFromUtf8(isolate, name, v8::String::kNormalString), sv2v8(thing));
 }
 
 void
 V8Context::bind_ro(const char *name, SV *thing) {
-    HandleScope scope;
-    Context::Scope context_scope(context);
+    Isolate::Scope isolate_scope(isolate);
+    HandleScope scope(isolate);
 
-    context->Global()->ForceSet(String::New(name), sv2v8(thing),
-        v8::PropertyAttribute(v8::ReadOnly | v8::DontDelete));
+    Local<Context> local_context = Local<Context>::New(isolate, context);
+    Context::Scope context_scope(local_context);
+
+    bool result = context.Get(isolate)->Global()->DefineOwnProperty(
+        isolate->GetCurrentContext(),
+        v8::String::NewFromUtf8(isolate, name, v8::String::kNormalString),
+        sv2v8(thing),
+        v8::PropertyAttribute(v8::ReadOnly | v8::DontDelete)
+    ).IsJust();
 }
 
 void V8Context::name_global(const char *name) {
-    HandleScope scope;
-    Context::Scope context_scope(context);
+    HandleScope scope(isolate);
 
-    context->Global()->ForceSet(String::New(name), context->Global(),
-        v8::PropertyAttribute(v8::ReadOnly | v8::DontDelete));
+    Local<Context> local_context = Local<Context>::New(isolate, context);
+    Context::Scope context_scope(local_context);
+
+    bool result = context.Get(isolate)->Global()->DefineOwnProperty(
+        isolate->GetCurrentContext(),
+        v8::String::NewFromUtf8(isolate, name, v8::String::kNormalString),
+        context.Get(isolate)->Global(),
+        v8::PropertyAttribute(v8::ReadOnly | v8::DontDelete)
+    ).IsJust();
 }
 
 // I fucking hate pthreads, this lacks error handling, but hopefully works.
@@ -426,7 +508,7 @@ private:
         ts.tv_nsec = tv.tv_usec * 1000;
 
         if (pthread_cond_timedwait(&me->cond_, &me->mutex_, &ts) == ETIMEDOUT) {
-            V8::TerminateExecution();
+            V8::TerminateExecution(isolate);
         }
         pthread_mutex_unlock(&me->mutex_);
         return NULL;
@@ -440,15 +522,17 @@ private:
 
 SV*
 V8Context::eval(SV* source, SV* origin) {
-    HandleScope handle_scope;
+    Isolate::Scope isolate_scope(isolate);
+    HandleScope handle_scope(isolate);
     TryCatch try_catch;
-    Context::Scope context_scope(context);
+    Local<Context> local_context = context.Get(isolate);
+    Context::Scope context_scope(local_context);
 
     // V8 expects everything in UTF-8, ensure SVs are upgraded.
     sv_utf8_upgrade(source);
     Handle<Script> script = Script::Compile(
         sv2v8str(source),
-        origin ? sv2v8str(origin) : String::New("eval")
+        origin ? sv2v8str(origin) : String::NewFromUtf8(isolate, "eval", v8::String::kNormalString)
     );
 
     if (try_catch.HasCaught()) {
@@ -478,23 +562,23 @@ V8Context::sv2v8(SV *sv, HandleMap& seen) {
     if (SvPOK(sv)) {
         // Upgrade string to UTF-8 if needed
         char *utf8 = SvPVutf8_nolen(sv);
-        return String::New(utf8, SvCUR(sv));
+        return String::NewFromUtf8(isolate, utf8, v8::String::kNormalString);
     }
     if (SvUOK(sv)) {
         UV v = SvUV(sv);
-        return (v < 0xffffffffUL) ? (Handle<Number>)Integer::NewFromUnsigned(v) : Number::New(SvNV(sv));
+        return (v < 0xffffffffUL) ? (Handle<Number>)Integer::NewFromUnsigned(isolate, v) : Number::New(isolate, SvNV(sv));
     }
     if (SvIOK(sv)) {
         IV v = SvIV(sv);
-        return (v <= INT32_MAX && v >= INT32_MIN) ? (Handle<Number>)Integer::New(v) : Number::New(SvNV(sv));
+        return (v <= INT32_MAX && v >= INT32_MIN) ? (Handle<Number>)Integer::New(isolate, v) : Number::New(isolate, SvNV(sv));
     }
     if (SvNOK(sv))
-        return Number::New(SvNV(sv));
+        return Number::New(isolate, SvNV(sv));
     if (!SvOK(sv))
-        return Undefined();
+        return Undefined(isolate);
 
     warn("Unknown sv type in sv2v8");
-    return Undefined();
+    return Undefined(isolate);
 }
 
 Handle<Value>
@@ -507,16 +591,26 @@ Handle<String> V8Context::sv2v8str(SV* sv)
 {
     // Upgrade string to UTF-8 if needed
     char *utf8 = SvPVutf8_nolen(sv);
-    return String::New(utf8, SvCUR(sv));
+    return String::NewFromUtf8(isolate, utf8, v8::String::kNormalString, SvCUR(sv));
 }
 
 SV* V8Context::seen_v8(Handle<Object> object) {
-    Handle<Value> wrap = object->GetHiddenValue(string_wrap);
-    if (wrap.IsEmpty())
-        return NULL;
+    v8::Local<v8::Private> privateKey = v8::Private::ForApi(isolate, string_wrap.Get(isolate));
+    v8::Local<v8::Value> value;
 
-    ObjectData* data = (ObjectData*)External::Unwrap(wrap);
-    return newRV(data->sv);
+    v8::Local<v8::Context> context = isolate->GetCurrentContext();
+
+    if (!object->HasPrivate(context, privateKey).ToChecked()) {
+        return NULL;
+    }
+
+    if (object->GetPrivate(context, privateKey).ToLocal(&value)) {
+        ObjectData* data = (ObjectData*) value.As<v8::External>()->Value();
+
+        return newRV(data->sv);
+    }
+
+    return NULL;
 }
 
 SV *
@@ -581,14 +675,14 @@ V8Context::v82sv(Handle<Value> value) {
 void
 V8Context::fill_prototype(Handle<Object> prototype, HV* stash) {
     HE *he;
-    while (he = hv_iternext(stash)) {
+    while ((he = hv_iternext(stash))) {
         SV *key = HeSVKEY_force(he);
-        Local<String> name = String::New(SvPV_nolen(key));
+        Local<String> name = v8::String::NewFromUtf8(isolate, SvPV_nolen(key), v8::String::kNormalString);
 
         if (prototype->Has(name))
             continue;
 
-        prototype->Set(name, (new PerlMethodData(this, SvPV_nolen(key)))->object);
+        prototype->Set(name, (new PerlMethodData(this, SvPV_nolen(key)))->object.Get(isolate));
     }
 }
 
@@ -601,25 +695,28 @@ V8Context::get_prototype(SV *sv) {
     std::string pkg(package);
     ObjectMap::iterator it;
 
-    Persistent<Object> prototype;
+    Persistent<Object, CopyablePersistentTraits<Object>> prototype;
 
     it = prototypes.find(pkg);
     if (it != prototypes.end()) {
         prototype = it->second;
     }
     else {
-        prototype = prototypes[pkg] = Persistent<Object>::New(Object::New());
+        Local<Object> object = Object::New(isolate);
+        Persistent<Object, CopyablePersistentTraits<Object>> persistent_object(isolate, object);
+        prototype = prototypes[pkg] = persistent_object;
+        //prototype = prototypes[pkg] = Persistent<Object>::New(isolate, Object::New(isolate));
 
         if (AV *isa = mro_get_linear_isa(stash)) {
             for (int i = 0; i <= av_len(isa); i++) {
                 SV **sv = av_fetch(isa, i, 0);
                 HV *stash = gv_stashsv(*sv, 0);
-                fill_prototype(prototype, stash);
+                fill_prototype(prototype.Get(isolate), stash);
             }
         }
     }
 
-    return prototype;
+    return prototype.Get(isolate);
 }
 #endif
 
@@ -631,7 +728,7 @@ V8Context::rv2v8(SV *rv, HandleMap& seen) {
     {
         ObjectDataMap::iterator it = seen_perl.find(ptr);
         if (it != seen_perl.end())
-            return it->second->object;
+            return it->second->object.Get(isolate);
     }
 
     {
@@ -646,7 +743,7 @@ V8Context::rv2v8(SV *rv, HandleMap& seen) {
         if ((0 == strcmp(Perl_class, "JSON::PP::Boolean"))
             || (0 == strcmp(Perl_class, "JSON::XS::Boolean"))
         )
-            return Boolean::New(sv_2bool(sv));
+            return Boolean::New(isolate, sv_2bool(sv));
         return blessed2object(sv);
     }
 #endif
@@ -663,27 +760,27 @@ V8Context::rv2v8(SV *rv, HandleMap& seen) {
         return cv2function((CV*)sv);
 
     warn("Unknown reference type in sv2v8()");
-    return Undefined();
+    return Undefined(isolate);
 }
 
 #if PERL_VERSION > 8
 Handle<Object>
 V8Context::blessed2object(SV *sv) {
-    Handle<Object> object = Object::New();
+    Handle<Object> object = Object::New(isolate);
     object->SetPrototype(get_prototype(sv));
 
-    return (new PerlObjectData(this, object, sv))->object;
+    return (new PerlObjectData(this, object, sv))->object.Get(isolate);
 }
 #endif
 
 Handle<Array>
 V8Context::av2array(AV *av, HandleMap& seen, long ptr) {
     I32 i, len = av_len(av) + 1;
-    Handle<Array> array = Array::New(len);
+    Handle<Array> array = Array::New(isolate, len);
     seen[ptr] = array;
     for (i = 0; i < len; i++) {
         if (SV** sv = av_fetch(av, i, 0)) {
-            array->Set(Integer::New(i), sv2v8(*sv, seen));
+            array->Set(Integer::New(isolate, i), sv2v8(*sv, seen));
         }
     }
     return array;
@@ -696,17 +793,17 @@ V8Context::hv2object(HV *hv, HandleMap& seen, long ptr) {
     SV *val;
 
     hv_iterinit(hv);
-    Handle<Object> object = Object::New();
+    Handle<Object> object = Object::New(isolate);
     seen[ptr] = object;
     while (val = hv_iternextsv(hv, &key, &len)) {
-        object->Set(String::New(key, len), sv2v8(val, seen));
+        object->Set(v8::String::NewFromUtf8(isolate, key, v8::String::kNormalString), sv2v8(val, seen));
     }
     return object;
 }
 
 Handle<Object>
 V8Context::cv2function(CV *cv) {
-    return (new PerlFunctionData(this, (SV*)cv))->object;
+    return (new PerlFunctionData(this, (SV*)cv))->object.Get(isolate);
 }
 
 SV*
@@ -717,7 +814,7 @@ V8Context::array2sv(Handle<Array> array, SvMap& seen) {
     seen.add(array, PTR2IV(av));
 
     for (int i = 0; i < array->Length(); i++) {
-        Handle<Value> elementVal = array->Get( Integer::New( i ) );
+        Handle<Value> elementVal = array->Get( Integer::New( isolate, i ) );
         av_push(av, v82sv(elementVal, seen));
     }
     return rv;
@@ -725,7 +822,7 @@ V8Context::array2sv(Handle<Array> array, SvMap& seen) {
 
 SV *
 V8Context::object2sv(Handle<Object> obj, SvMap& seen) {
-    if (enable_blessing && obj->Has(String::New("__perlPackage"))) {
+    if (enable_blessing && obj->Has(v8::String::NewFromUtf8(isolate, "__perlPackage", v8::String::kNormalString))) {
         return object2blessed(obj);
     }
 
@@ -736,7 +833,7 @@ V8Context::object2sv(Handle<Object> obj, SvMap& seen) {
 
     Local<Array> properties = obj->GetPropertyNames();
     for (int i = 0; i < properties->Length(); i++) {
-        Local<Integer> propertyIndex = Integer::New( i );
+        Local<Integer> propertyIndex = Integer::New( isolate, i );
         Local<String> propertyName = Local<String>::Cast( properties->Get( propertyIndex ) );
         String::Utf8Value propertyNameUTF8( propertyName );
 
@@ -771,14 +868,15 @@ my_gv_setsv(pTHX_ GV* const gv, SV* const sv){
 \
     { \
         /* We have to do all this inside a block so that all the proper \
-         * destuctors are called if we need to croak. If we just croak in the \
+         * destructors are called if we need to croak. If we just croak in the \
          * middle of the block, v8 will segfault at program exit. */ \
+        Isolate::Scope  isolate_scope(isolate); \
+        HandleScope     scope(isolate); \
         TryCatch        try_catch; \
-        HandleScope     scope; \
         V8FunctionData* data = (V8FunctionData*)sv_object_data((SV*)cv); \
         if (data->context) { \
         V8Context      *self = data->context; \
-        Handle<Context> ctx  = self->context; \
+        Handle<Context> ctx  = self->context.Get(isolate); \
         Context::Scope  context_scope(ctx); \
         vector<Handle<Value> > argv; \
 \
@@ -798,7 +896,7 @@ my_gv_setsv(pTHX_ GV* const gv, SV* const sv){
                     count = array->Length(); \
                     EXTEND(SP, count - items); \
                     for (int i = 0; i < count; i++) { \
-                        ST(i) = sv_2mortal(self->v82sv(array->Get(Integer::New(i)))); \
+                        ST(i) = sv_2mortal(self->v82sv(array->Get(Integer::New(isolate, i)))); \
                     } \
                 } \
                 else { \
@@ -824,14 +922,14 @@ XSRETURN(count);
 
 XS(v8closure) {
     SETUP_V8_CALL(0)
-    Handle<Value> result = Handle<Function>::Cast(data->object)->Call(ctx->Global(), items, &argv[0]);
+    Handle<Value> result = Handle<Function>::Cast(data->object.Get(isolate))->Call(ctx->Global(), items, &argv[0]);
     CONVERT_V8_RESULT()
 }
 
 XS(v8method) {
     SETUP_V8_CALL(1)
     V8ObjectData* This = (V8ObjectData*)SvIV((SV*)SvRV(ST(0)));
-    Handle<Value> result = Handle<Function>::Cast(data->object)->Call(This->object, items - 1, &argv[0]);
+    Handle<Value> result = Handle<Function>::Cast(data->object.Get(isolate))->Call(This->object.Get(isolate), items - 1, &argv[0]);
     CONVERT_V8_RESULT(POPs);
 }
 
@@ -846,12 +944,14 @@ SV*
 V8Context::object2blessed(Handle<Object> obj) {
     char package[128];
 
+    String::Utf8Value stringified(obj->Get( String::NewFromUtf8(isolate, "__perlPackage") )->ToString());
+
     snprintf(
         package,
         128,
         "%s%s::N%d",
         bless_prefix.c_str(),
-        *String::AsciiValue(obj->Get(String::New("__perlPackage"))->ToString()),
+        *stringified,
         number
     );
 
@@ -875,8 +975,8 @@ V8Context::object2blessed(Handle<Object> obj) {
             CV *code = newXS(NULL, v8method, __FILE__);
             V8ObjectData *data = new V8FunctionData(this, fn, (SV*)code);
 
-            GV* gv = (GV*)*hv_fetch(stash, *String::AsciiValue(name), name->Length(), TRUE);
-            gv_init(gv, stash, *String::AsciiValue(name), name->Length(), GV_ADDMULTI); /* vivify */
+            GV* gv = (GV*)*hv_fetch(stash, *String::Utf8Value(name), name->Length(), TRUE);
+            gv_init(gv, stash, *String::Utf8Value(name), name->Length(), GV_ADDMULTI); /* vivify */
             my_gv_setsv(aTHX_ gv, (SV*)code);
         }
     }
@@ -901,12 +1001,15 @@ V8Context::idle_notification() {
         hs.used_heap_size()
     );
     */
-    return V8::IdleNotification();
+
+    isolate->LowMemoryNotification(); // force garbage collection
+
+    return true;
 }
 
 int
 V8Context::adjust_amount_of_external_allocated_memory(int change_in_bytes) {
-    return V8::AdjustAmountOfExternalAllocatedMemory(change_in_bytes);
+    //return V8::AdjustAmountOfExternalAllocatedMemory(change_in_bytes);
 }
 
 void
